# -*- coding: utf-8 -*-
from django.core.cache import caches
from django.db.models.signals import post_delete
from django.dispatch import receiver
from oauth2_provider.models import AccessToken

from django_toolkit import toolkit_settings

cache = caches[toolkit_settings.ACCESS_TOKEN_CACHE_BACKEND]


@receiver(post_delete, sender=AccessToken)
def invalidate_token_cache(sender, instance, **kwargs):
    cache.delete(instance.token)



# -*- coding: utf-8 -*-
import logging
import socket


class AddHostName(logging.Filter):
    """
    Add the %(hostname)s entry to log record, so that it can be included in the
    logs using a formatter
    """
    hostname = socket.gethostname()

    def filter(self, record):
        record.hostname = self.hostname
        return True



# -*- coding: utf-8 -*-
import logging

logger = logging.getLogger(__name__)


class CircuitBreaker:

    def __init__(
        self,
        cache,
        failure_cache_key,
        max_failures,
        max_failure_exception,
        max_failure_timeout=None,
        circuit_timeout=None,
        catch_exceptions=None,
    ):
        self.cache = cache
        self.failure_cache_key = failure_cache_key
        self.max_failure_timeout = max_failure_timeout
        self.circuit_timeout = circuit_timeout
        self.circuit_cache_key = 'circuit_{}'.format(failure_cache_key)
        self.max_failure_exception = max_failure_exception
        self.catch_exceptions = catch_exceptions or (Exception,)
        self.max_failures = max_failures

    @property
    def is_circuit_open(self):
        return self.cache.get(self.circuit_cache_key) or False

    @property
    def total_failures(self):
        return self.cache.get(self.failure_cache_key) or 0

    def open_circuit(self):
        logger.critical(
            'Open circuit for {failure_cache_key} {cicuit_cache_key}'.format(
                failure_cache_key=self.failure_cache_key,
                cicuit_cache_key=self.circuit_cache_key
            )
        )

        self.cache.set(self.circuit_cache_key, True, self.circuit_timeout)

    def __enter__(self):
        if self.is_circuit_open:
            raise self.max_failure_exception

    def __exit__(self, exc_type, exc_value, traceback):
        if exc_type in self.catch_exceptions:

            self._increase_failure_count()

            if self.total_failures >= self.max_failures:
                self.open_circuit()

                logger.info(
                    'Max failures exceeded by: {}'.format(
                        self.failure_cache_key
                    )
                )

                raise self.max_failure_exception

    def _increase_failure_count(self):
        self.cache.add(self.failure_cache_key, 0, self.max_failure_timeout)
        total = self.cache.incr(self.failure_cache_key)

        logger.info(
            'Increase failure for: {key} - '
            'max failures {max_failures} - '
            'total {total}'.format(
                key=self.failure_cache_key,
                max_failures=self.max_failures,
                total=total
            )
        )



# -*- coding: utf-8 -*-
import os

from setuptools import find_packages, setup


def read(fname):
    path = os.path.join(os.path.dirname(__file__), fname)
    with open(path) as f:
        return f.read()

setup(
    name='luizalabs-django-toolkit',
    version='1.3.0',
    description=(
        'The LuizaLabs set of tools '
        'to develop projects using the Django framework'
    ),
    long_description=read('README.rst'),
    author='Luizalabs',
    url='https://github.com/luizalabs/django-toolkit',
    keywords='django tools logs middleware utils',
    install_requires=[
        'Django>=1.8',
    ],
    extra_requires={
        'oauth2': ['django-oauth-toolkit'],
    },
    packages=find_packages(exclude=[
        'tests*'
    ]),
    classifiers=[
        'Framework :: Django',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 3',
    ]
)



# -*- coding: utf-8 -*-
import logging

from .shortcuts import get_oauth2_app
from .toolkit_settings import API_VERSION, MIDDLEWARE_ACCESS_LOG_FORMAT

try:
    from django.utils.deprecation import MiddlewareMixin
except ImportError:
    MiddlewareMixin = object


logger = logging.getLogger(__name__)


class VersionHeaderMiddleware(MiddlewareMixin):
    """
    Add a X-API-Version header to the response. The version is taken from
    TOOLKIT['API_VERSION'] setting.
    """

    def process_response(self, request, response):
        response['X-API-Version'] = API_VERSION
        return response


class AccessLogMiddleware(MiddlewareMixin):

    LOG_FORMAT = MIDDLEWARE_ACCESS_LOG_FORMAT
    UNKNOWN_APP_NAME = 'unknown'

    def process_response(self, request, response):
        app = get_oauth2_app(request)

        app_name = getattr(app, 'name', self.UNKNOWN_APP_NAME)

        logger.info(
            self.LOG_FORMAT.format(app_name=app_name,
                                   request=request,
                                   response=response)
        )

        return response



from django.core.cache import caches


class LockActiveError(Exception):
    pass


class Lock(object):

    def __init__(self):
        self.active = False


class LocalMemoryLock(Lock):
    """
    A context manager to handle a lock status in memory

    The active status is True on __enter__ and False on __exit__
    """

    def __enter__(self):

        if self.active:
            raise LockActiveError('Lock is already active')

        self.active = True
        return self

    def __exit__(self, *args, **kwargs):
        self.active = False


class CacheLock(Lock):
    """
    A context manager to handle a lock status using Django cache

    The active status is True on __enter__ and False on __exit__

    The cache will be deleted on context __exit__
    """

    def __init__(
        self,
        key,
        cache_alias='default',
        expire=0,
        raise_exception=True
    ):
        super(CacheLock, self).__init__()
        self._key = key
        self._expire = expire
        self.cache = caches[cache_alias]
        self.raise_exception = raise_exception

    def __enter__(self):
        self.active = self.cache.add(self._key, True, self._expire)

        if not self.active and self.raise_exception:
            raise LockActiveError('For key {key}'.format(key=self._key))

        return self

    def __exit__(self, *args, **kwargs):
        self.active = False
        self.cache.delete(self._key)



# -*- coding: utf-8 -*-
from django.apps import AppConfig


class OAuth2AppConfig(AppConfig):
    name = 'django_toolkit.oauth2'
    verbose_name = 'OAuth2'

    def ready(self):
        super(OAuth2AppConfig, self).ready()
        from . import receivers  # noqa



# -*- coding: utf-8 -*-
def get_oauth2_app(request):
    """
    Return the Application object from a Django Rest Framework request when
    authenticated with a Django OAuth Toolkit access token, or return None
    when it is not
    """
    try:
        return request.auth.application
    except AttributeError:
        return None



# -*- coding: utf-8 -*-
from django.core.cache import caches
from django.utils import timezone
from oauth2_provider.models import AccessToken
from oauth2_provider.oauth2_validators import OAuth2Validator

from django_toolkit import toolkit_settings

cache = caches[toolkit_settings.ACCESS_TOKEN_CACHE_BACKEND]


class CachedOAuth2Validator(OAuth2Validator):

    def validate_bearer_token(self, token, scopes, request):
        if not token:
            return False

        try:
            access_token = self._get_access_token(token)
            if access_token.is_valid(scopes):
                request.client = access_token.application
                request.user = access_token.user
                request.scopes = scopes

                # this is needed by django rest framework
                request.access_token = access_token
                return True
            return False
        except AccessToken.DoesNotExist:
            return False

    def _get_access_token(self, token):
        access_token = cache.get(token)

        if access_token is None:
            access_token = AccessToken.objects.select_related(
                'application',
                'user'
            ).get(
                token=token
            )
            now = timezone.now()
            if (access_token.expires > now):
                timeout = (access_token.expires - now).seconds

                cache.set(token, access_token, timeout)

        return access_token



from django.conf import settings

_toolkit_settings = getattr(settings, 'TOOLKIT', {})

ACCESS_TOKEN_CACHE_BACKEND = _toolkit_settings.get(
    'ACCESS_TOKEN_CACHE_BACKEND',
    'access_token'
)

API_VERSION = _toolkit_settings.get('API_VERSION')

MIDDLEWARE_ACCESS_LOG_FORMAT = _toolkit_settings.get(
    'MIDDLEWARE_ACCESS_LOG_FORMAT',
    u'[{app_name}] {response.status_code} {request.method} {request.path}'
)
